package bank;

import javax.ejb.EJBObject;
import java.rmi.RemoteException;
import java.util.List;

/**
 * @author S.Chassande-Barrioz
 */
public interface BankApplication extends EJBObject {

    // Agencies management
    //--------------------

    /**
     * Create a new Bank in the bank
     * @param name is the name of the agency
     */
    void createAgency(String name) throws RemoteException;

    /**
     * Retrieves a list of agency names.
     * @return list of String
     */
    List getAgencies() throws RemoteException;

    /**
     * Removes an agency only if this agency does not have any client
     * @param name is the name of the agency to remove
     * @return true if the agency was removed
     */
    boolean removeAgency(String name) throws RemoteException;

    
    // Clients management
    //--------------------

    /**
     * Create a Client in an agency
     * @param cid is a client identifier (last name, first name and address)
     * @param agencyName the the name of the agency of the new client
     */
    void createClient(ClientId cid, String agencyName) throws RemoteException;

    /**
     * Retrieves a list of ClientId instance corresponding to the clients of
     * an agency.
     * @param agencyName is the name of the agency
     * @return list of ClientId
     */
    List getClients(String agencyName) throws RemoteException;

    /**
     * Removes a client from an agency only if the client does not have any Account
     * @param agencyName is the name of the agency hosting the client
     * @param cid is a client identifier
     * @return true if the client was removed
     */
    boolean removeClient(String agencyName, ClientId cid) throws RemoteException;


    // Accounts management
    //--------------------

    /**
     * Create a new Account for a client of an agency
     * @param cid is a client identifier (last name, first name and address),
     * the owner of the new account.
     * @param agencyName the name of the agency of the client
     * @return the account number
     */
    String createAccount(ClientId cid, String agencyName) throws RemoteException;

    /**
     * Retrieves information about an account of a client in an agency
     * @param number is the account number
     * @param cid is the client identifier
     * @param agencyName is the name of the agency
     */
    AccountInfo getAccountInfo(String number,
                               ClientId cid,
                               String agencyName) throws RemoteException;

    /**
     * Retrieves a list of account for client.
     * @param agencyName is the name of the agency
     * @param cid is the client identifier
     * @return list of AccountInfo instance
     */
    List getAccounts(String agencyName, ClientId cid) throws RemoteException;

    /**
     * Closes a account of a client.
     * @param agencyName is the agency name
     * @param cid is a client identifier
     * @param accountNumber is the number of account to close
     * @return a AccountInfo if the account was removed. The accountInfo
     * instance contains the value of the account. If the account does not exist
     * a null value is returned
     */
    AccountInfo closeAccount(String agencyName,
                             ClientId cid,
                             String accountNumber) throws RemoteException;
}
