/**
 * Copyright (C) 2001-2005 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.genclass;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Collection;


import org.objectweb.jorm.api.PExceptionIO;
import org.objectweb.jorm.api.PGenClassAccessor;
import org.objectweb.jorm.api.PIndexedElem;
import org.objectweb.jorm.util.api.Loggable;
import org.objectweb.speedo.api.Debug;
import org.objectweb.speedo.genclass.api.SpeedoGenClassCoherence;
import org.objectweb.speedo.genclass.api.SpeedoGenClassPO;
import org.objectweb.speedo.mim.api.StateItf;
import org.objectweb.speedo.mim.lib.AbstractStateImpl;
import org.objectweb.speedo.pm.api.POManagerItf;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * This class represents an accessor of a genclass structure. This accessor is
 * used in case of the generic class is persistent.
 * 
 * @author S.Chassande-Barrioz
 */
public abstract class GenClassAccessor
	extends AbstractStateImpl
	implements PGenClassAccessor, SpeedoGenClassCoherence {

	/**
	 * The PersistentObjectItf attached to this state representation.
	 */
	public SpeedoGenClassPO gcpo;

	/**
	 * Indicates if this accessor support the dela mode
	 */
	public boolean supportDelta = true;

	/**
	 * A temporary variable which is used to search an element into the
	 * "elements" list.
	 */
	protected GenClassElement tmpelem = null;

	protected transient Logger logger;

	/**
	 * The ArrayList used to store the indexed elements of the genclass. Then
	 * this list contains also the deleted elements.
	 */
	public List elements = null;
	
	public List deltaForMerge = null;

	public GenClassAccessor(SpeedoGenClassPO thepo) {
        super(thepo);
		elements = new ArrayList();
		this.gcpo = thepo;
		this.tmpelem = (GenClassElement) createPIndexedElem();
	}

	public abstract void loadFieldsFromAccessor(StateItf sa);
	
	public abstract PIndexedElem createPIndexedElem(GenClassAccessor gca);

	public abstract void setElements(Object o);

	public Logger getLogger() {
		if (logger == null) {
			logger = ((Loggable) gcpo).getLogger();
		}
		return logger;
	}

	/**
	 * When the gen class is flushed to the data support (parameter = true),
	 * the internal structure is clean: - really remove the deleted element -
	 * mark the genclass as supporting the delta
	 */
	public void setFlushed(boolean val) {
		super.setFlushed(val);
		if (val == true) {
			supportDelta = true;
			Iterator it = elements.iterator();
			while (it.hasNext()) {
				GenClassElement gce = (GenClassElement) it.next();
				if (gce.getElemStatus() == PIndexedElem.ELEM_DELETED) {
					//Remove the deleted element
					it.remove();
				}
			}
		}
	}

	/**
	 * At the end of the working set, the writing mode is initialized
	 * (deltaSupported), and the reference to the persistent are unswilled
	 */
	public void workingSetClosed() {
		if (Debug.ON) {
			((Loggable) gcpo).getLogger().log(
			        BasicLevel.DEBUG, "workingSetClosed()");
		}
		supportDelta = true;
		for(Iterator it = elements.iterator(); it.hasNext();) {
			GenClassElement gce = (GenClassElement) it.next();
			if (gce.getElemStatus() == PIndexedElem.ELEM_DELETED) {
				//Remove the deleted element
				it.remove();
			}
		}
	}

	public void unSwizzle() {
		for(Iterator it = elements.iterator(); it.hasNext();) {
			GenClassElement gce = (GenClassElement) it.next();
			gce.unSwizzle();
		}
	}

	public org.objectweb.perseus.persistence.api.State merge(org.objectweb.perseus.persistence.api.State oldState) {
        if (deltaForMerge == null || !isToMerge) {
            return super.merge(oldState);
        }
        GenClassAccessor gca  = (GenClassAccessor) oldState;
        for(int i=0; i<deltaForMerge.size(); i++) {
            GenClassElement gce = (GenClassElement) deltaForMerge.get(i);
            switch(gce.getStatusForMerge()) {
            case PIndexedElem.ELEM_CREATED:
                gca.elements.add(gce);
                break;
            case PIndexedElem.ELEM_DELETED:
                int idx = gca.elements.indexOf(gce);
            	if (idx == -1) {
            	    logger.log(BasicLevel.WARN, 
            	            "Deleted GenClass element '" + gce + "' not found during merging among " 
            	            + gca.elements);
            	} else {
            	    gca.elements.remove(idx);
            	}
                break;
            case PIndexedElem.ELEM_MODIFIED:
                idx = gca.elements.indexOf(gce);
            	if (idx == -1) {
            	    logger.log(BasicLevel.WARN, 
        	            "Modified GenClass element not found during merging: " 
        	            + gce);
            	} else {
            	    gca.elements.set(idx, gce);
            	}
                break;
            }
            gce.cleanStatusForMerge();
        }
        deltaForMerge = null;
        return super.merge(oldState);
    }
    
    public void makeToMerge(Object thinLock) {
        if (!isToMerge) {
            for(int i=0; i<elements.size(); i++) {
                elements.set(i, ((GenClassElement) elements.get(i)).cloneGCE());
            }
        }
        super.makeToMerge(thinLock);
    }
    
	public void loadFields(POManagerItf pm, long[] fields) {
	}

	public abstract void deletePersistent(POManagerItf pm);

	public abstract void makePersistent(POManagerItf pm);
	
	public abstract void makePersistentOnAttach(POManagerItf pm, Map map);

	public abstract void detachCopy(POManagerItf pm, Map map, StateItf fieldsClone,  Collection fgHints);
	
	public abstract void attachCopy(POManagerItf pm, Map map, StateItf fieldsClone);
	
	public abstract void refresh(POManagerItf pm, Map map, Collection fgHints);
	
	public abstract void retrieve(POManagerItf pm, Map map, Collection fgHints);
	
	// IMPLEMENTATION OF THE SpeedoGenClassCoherence INTERFACE //
	//---------------------------------------------------------//

	public abstract boolean speedoAdd(Object elemToAdd, Object hints);

	public abstract boolean speedoRemove(Object elemToRemove, Object hints);

	// IMPLEMENTATION OF THE PGenClassAccessor INTERFACE //
	//---------------------------------------------------//

	public PIndexedElem createPIndexedElem() {
		return createPIndexedElem(this);
	}

	public Object getMemoryInstance() {
		return gcpo;
	}

	public void paAdd(PIndexedElem elem, Object conn) throws PExceptionIO {
		// the elem is read from the DS, set it to unmodified.
		 ((GenClassElement) elem).setStatus(PIndexedElem.ELEM_UNMODIFIED);
		elements.add(elem);
	}

	public int paGetNbElem() {
		return elements.size();
	}

	public Iterator paIterator() {
		return elements.iterator();
	}

	public boolean paDeltaSupported() {
		return supportDelta;
	}

	public void paSetNbElem(int nbelem) {
		elements = (nbelem < 0 ? new ArrayList() : new ArrayList(nbelem));
	}

}
