/**
 * Speedo: an implementation of JDO compliant personality on top of JORM generic
 * I/O sub-system.
 * Copyright (C) 2001-2005 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *
 * Contact: speedo@objectweb.org
 *
 * Authors: S.Chassande-Barrioz.
 *
 */

package org.objectweb.speedo.mim.api;

import java.util.Collection;

import javax.jdo.Extent;
import javax.jdo.PersistenceManager;
import javax.jdo.Query;

/**
 * Fetch groups are activated using methods on this interface. An
 * instance of this interface can be obtained from {@link
 * PersistenceManager#getFetchPlan}, {@link Extent#getFetchPlan}, and
 * {@link Query#getFetchPlan}. When a <code>Query</code> or
 * <code>Extent</code> is retrieved from a
 * <code>PersistenceManager</code>, its <code>FetchPlan</code> is
 * initialized to the same settings as that of the
 * <code>PersistenceManager</code>. Subsequent modifications of the
 * <code>Query</code> or <code>Extent</code>'s <code>FetchPlan</code>
 * are not reflected in the <code>FetchPlan</code> of the
 * <code>PersistenceManager</code>.
 */
public interface FetchPlanItf {
    /**
     * For use with {@link #speedoAddGroup(String)}, {@link #speedoRemoveGroup(String)}, and the
     * various {@link #speedoSetGroups} calls. Value: <code>default</code>.
     */
    public static final String DEFAULT = "default";

    /**
     * For use with {@link #speedoAddGroup(String)}, {@link #speedoRemoveGroup(String)}, and the
     * various {@link #speedoSetGroups} calls. Value: <code>all</code>.
     */
    public static final String ALL = "all";

    /**
     * For use with {@link #speedoAddGroup(String)}, {@link #speedoRemoveGroup(String)}, and the
     * various {@link #speedoSetGroups} calls. Value: <code>values</code>.
     */
    public static final String VALUES = "values";

    /**
     * For use with {@link #speedoAddGroup(String)}, {@link #speedoRemoveGroup(String)}, and the
     * various {@link #speedoSetGroups} calls. Value: <code>none</code>.
     */
    public static final String NONE = "none";

    /**
     * For use with {@link #setFetchSize}. Value: -1.
     */
    public static final int FETCH_SIZE_GREEDY = -1;

    /**
     * For use with {@link #setFetchSize}. Value: 0.
     */
    public static final int FETCH_SIZE_OPTIMAL = 0;

    /** 
     * Add the fetch group to the set of active fetch groups.
     * @return the FetchPlan
     */
    FetchPlanItf speedoAddGroup(String fetchGroupName);

    /** 
     * Remove the fetch group from the set active fetch groups. 
     * @return the FetchPlan
     */
    FetchPlanItf speedoRemoveGroup(String fetchGroupName);

    /** 
     * Remove all active groups leaving no active fetch group.
     * @return the FetchPlan
     */ 
    FetchPlanItf speedoClearGroups();

    /** 
     * Return the names of all active fetch groups.
     * @return the names of active fetch groups
     * @return the FetchPlan
     */
    Collection speedoGetGroups();

    /** 
     * Set a collection of groups.
     * @param fetchGroupNames a collection of names of fetch groups
     * @return the FetchPlan
     */
    FetchPlanItf speedoSetGroups(Collection fetchGroupNames);

    /** 
     * Set a collection of groups.
     * @param fetchGroupNames a String array of names of fetch groups
     * @return the FetchPlan
     */
    FetchPlanItf speedoSetGroups(String[]fetchGroupNames);

    /** 
     * Set the active fetch groups to the single named fetch group.
     * @param fetchGroupName the single fetch group
     * @return the FetchPlan
     */
    FetchPlanItf speedoSetGroup(String fetchGroupName);

    /**
     * Set the fetch size for large result set support. Use
     * {@link #FETCH_SIZE_OPTIMAL} to unset, and {@link #FETCH_SIZE_GREEDY}
     * to force loading of everything.
     * @param fetchSize the fetch size
     * @return the FetchPlan
     */
    FetchPlanItf speedoSetFetchSize(int fetchSize);

    /**
     * Return the fetch size, or {@link #FETCH_SIZE_OPTIMAL} if not set,
     * or {@link #FETCH_SIZE_GREEDY} to fetch all.
     * @return the fetch size
     */
    int speedoGetFetchSize(); 

}
