/**
 * Speedo: an implementation of JDO compliant personality on top of JORM generic
 * I/O sub-system.
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *
 * Contact: speedo@objectweb.org
 *
 * Authors: S.Chassande-Barrioz.
 *
 */

package org.objectweb.speedo.mim.api;

import org.objectweb.speedo.pm.api.POManagerItf;
import org.objectweb.jorm.api.PAccessor;
import java.io.Serializable;
import java.util.Collection;
import java.util.Map;

/**
 * This interface a state of an user instance.
 *
 * @author S.Chassande-Barrioz
 */
public interface StateItf extends PAccessor, org.objectweb.perseus.persistence.api.State, Serializable {

    /**
     * Changes the status of this object
     *
     * @see org.objectweb.speedo.mim.api.LifeCycle
     * @param action the action that may change the status
     */
    void speedoChangeStatus(byte action);

    /**
     * Gets the current status of this object
     *
     * @see org.objectweb.speedo.mim.api.LifeCycle
     * @return the current state in the life cycle
     */
    byte speedoGetStatus();

    /**
     * Forces the new value of the status
     *
     * @see org.objectweb.speedo.mim.api.LifeCycle
     * @param newValue the new status of this object
     */
    void speedoSetStatus(byte newValue);

	/**
	 * Version management:
	 *  strategies supported are:
	 * 		- date-time
	 * 		- version-number
	 * 
	 */
	
	/**
	 * change the version of this only if needed
	 * The version is computed by this method
	 * A call to changeVersion can have no effect.
	 */
	void speedoChangeVersion();
	
	/**
	 * Return the version of the object as a long
	 */
	long getVersion();
	
	/**
	 * check the validity of the version of sa compared to the persistent object this.
	 */
	boolean checkVersion(StateItf sa); 
	
	/**
	 * Gets the current status of this detached object
	 *
	 * @return the current state in the life cycle
	 */
	byte getDetachedStatus();

	/**
	 * Forces the new value of the detached status
	 *
	 * @param newValue the new status of this object
	 */
	void setDetachedStatus(byte newValue);
	  
    /**
     * @return The PersistentObjectItf attached to this state representation.
     */
    PersistentObjectItf getSpeedoPO();

    /**
     * Assignes the PersistentObjectItf attached to this state representation.
     */
    void setSpeedoPO(PersistentObjectItf sp);

    /**
     * Prepare to write (relations consistency management)
     */
    void prepareWrite();

    /**
     * The context has been closed (ex: transaction end).
     */
    void workingSetClosed();

	/**
	 * indicates if the object has been flushed on the support in a writing
	 * before the end of the working set (commit | rollback | close).
	 */
	boolean hasBeenFlush();

	/**
	 * Assignates a boolean value to indicate if the object has been flushed
	 * on the support in a writing before the end of the working
	 * set (commit | rollback | close).
	 * @param val is the boolean which must be assigned.
	 */
	void setFlushed(boolean val); 
	
	/**
	 * Makes persistent all inner elements of the persistent instance. For a 
	 * class, it means making persistent all objects referenced by a persistent
	 * field. For a genclass it means making persistent element of the genclass.
	 * This method is called when a makePersistent is called on the persistent
	 * class (PO) linked to this accessor.  
	 * @param pm is the po manager for making persistent inner element.
	 */
	void makePersistent(POManagerItf pm);
	
	/**
	 * makes persistent on attach, taking into account the map
	 * representing the attch context.
	 * It enables to avoid cycles. 
	 * @param pm is the po manager for making persistent inner element.
	 * @param map is the context of the attachment process.
	 */
	void makePersistentOnAttach(POManagerItf pm, Map map);
	
	/**
	 * Deletes persistent all inner elements of the persistent instance. For a 
	 * class, it means deleting persistent all objects referenced by a persistent
	 * field. For a genclass it means deleting persistent element of the genclass.
	 * This method is called when a deletePersistent is called on the persistent
	 * class (po) linked to this accessor.  
	 * @param pm is the po manager for making persistent inner element.
	 */
	void deletePersistent(POManagerItf pm);
	
	/**
	 * Askes the loading of the some fields identified by the long array.
	 * @param pm is the po manager to use to load references
	 * @param fields is an array of long. Each long is a bit mask indicating
	 * fields to load. As a long is stored over 64 bits, each long of the array
	 * can represent only 64 persistent fields. If the class has 70 persistent
	 * field, the parameter could be a long array size is 2. This parameter
	 * can be null if no field loading is required.
	 */
	void loadFields(POManagerItf pm, long[] fields);
	
	/**
	 * This method is used during the detach operation. A call to this method
	 * copy the field of the state into the the cloned state. When a field 
	 * references a persistent object, according to the fetch plan, the 
	 * referenced can be detached too.
	 * @param pm is the po manager used during the detach operation
	 * @param map is the map of the detached object. The use of this map permits
	 * to support cycle in the object graph.
	 * @param clone is a clone of the current state
	 * @param fgHints represents the fetch plan.
	 */
	void detachCopy(POManagerItf pm, 
	        Map map, 
	        StateItf clone, 
	        Collection fgHints);

	/**
	 * This method is used during the attach operation. A call to this method
	 * copy the field of the cloned state into the the current state. When a field 
	 * references a persistent object, according to the fetch plan, the 
	 * referenced can be attached too.
	 * @param pm is the po manager used during the attach operation
	 * @param map is the map of the atttached object. The use of this map permits
	 * to support cycle in the object graph.
	 * @param detached is the detached state used to update the current state
	 */
	void attachCopy(POManagerItf pm, 
	        Map map, 
	        StateItf detached);
	
	/**
	 * This method is used during the refresh operation. A call to this method
	 * refreshes the fields of the current state. When a field 
	 * references a persistent object, according to the fetch plan, the 
	 * referenced can be refreshed too.
	 * @param pm is the po manager used during the refresh operation
	 * @param map is the map of the refreshed object. The use of this map permits
	 * to support cycle in the object graph.
	 * @param fgHints represents the fetch plan.
	 */
	void refresh(POManagerItf pm, Map map, Collection fgHints);
	
	/**
	 * This method is used during the retrieve operation. A call to this method
	 * retrieves the fields of the current state. When a field 
	 * references a persistent object, according to the fetch plan, the 
	 * referenced can be retrieved too.
	 * @param pm is the po manager used during the retrieve operation
	 * @param map is the map of the retrieved object. The use of this map permits
	 * to support cycle in the object graph.
	 * @param fgHints represents the fetch plan.
	 */
	void retrieve(POManagerItf pm, Map map, Collection fgHints);
	
	/**
	 * Force the detached status to DETACHED_DIRTY and
	 * all the xxxLoaded to true
	 */
	void forceDetachedDirty();
	
	/**
	 * Restore the detached status to DETACHED_NONE and
	 * all the xxxLoaded to false
	 */
	void restoreDetachedNone();
	
    /**
     * indicates if the StateItf has to be merge with the reference state at commit
     * time.
     */
	boolean isToMerge();
    
    /**
     * Mark the current state with the flag 'toMerge'. This status means at 
     * commit time this state has to be merge with the referenceState (cache).
     * Since this method call and until the transaction end, the implementation
     * could register the change for the merge.
     * 
     * @param thinLock is an object representing the thin lock
     */
	void makeToMerge(Object thinLock);
    
    /**
     * This action updates the old state with the change (delta) done in the
     * transaction on this state.
     * @param oldState is the state which has to be updated
     * @return the updated state.
     */
    org.objectweb.perseus.persistence.api.State merge(org.objectweb.perseus.persistence.api.State oldState);
	
	void unSwizzle();
	
	/**
	 * The field uses in user cache, has been modified.
	 * @param cacheId is the cache identifier. The value Integer.MAX_VALUE
	 * means that all user cache must be updated.
	 * @param rebind if true, the speedoAcessor is rebound
	 */
	void indexFieldModified(int cacheId, boolean rebind);
}

