/**
 * Copyright (C) 2001-2006 France Telecom R&D
 */
package org.objectweb.speedo.naming.api;

import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.naming.api.PName;
import org.objectweb.jorm.naming.api.PNameCoder;
import org.objectweb.speedo.api.SpeedoException;
import org.objectweb.speedo.mapper.api.JormFactory;
import org.objectweb.speedo.metadata.SpeedoClass;

/**
 * A NamingManagerFactory manages a set of NamingManager instance. It permits 
 * - To bind or unbind a NamingManager,
 * - To find the right NamingManager managing the identifier/naming of a 
 * persistent class,
 * - To encode or decode persistent identifier.
 * 
 * @see NamingManager
 * @author S.Chassande-Barrioz
 */
public interface NamingManagerFactoryItf {

    /**
     * Binds a new NamaingManager in this factory
     * @param nm is the new NamingManager instance
     */
    void bindNamingManager(NamingManager nm);
    /**
     * Unbinds an existing NamaingManager in this factory
     * @param nm is the NamingManager instance to forget
     * @return true is the NamingManager has been really found and therefore 
     * forgoten.
     */
    boolean unbindNamingManager(NamingManager nm);
    /**
     * @return the NamingManager instance able to manage the identifier of a 
     * persistent class.
     * @param sc is the speedo meta object representing the persistent class.
     * @throws SpeedoException if no NamingManager can manage the persistent 
     * class.
     */
    NamingManager getNamingManager(SpeedoClass sc) throws SpeedoException;
    /**
     * @return the NamingManager instance able to manage the identifier of a 
     * persistent class.
     * @param hints is string value containing information permitting to find
     * the naming manager required.
     * @param classloader is the class loader of the application.
     * @throws SpeedoException if no NamingManager can manage the persistent 
     * class.
     */
    NamingManager getNamingManager(String hints, 
            ClassLoader classloader) throws PException;
    /**
     * Decodes an object identifer into a PName.
     * @param pnc is the PNamingContext managing the naming of the referenced 
     * persistent class
     * @param oid is the object identifier
     * @param clazz is the java class of the referenced class. this paramter 
     * permits of course to access to the class loader of the application. 
     * @param jf is the JormFactory managing the persistent class.
     * @return a PName (JORM identifier of persistent class)
     * @throws PException if it is not possible to decode the object identifier
     * corresponding to the specified class.
     * @see #encode(PName)
     */
    PName decode(PNameCoder pnc, Object oid, 
            java.lang.Class clazz, 
            JormFactory jf) throws PException;
    /**
     * Encodes the persistent identifier (PName) into an object identifier 
     * usable, externaly to the Speedo system.
     * @param pn is the Pname to encode
     * @see #decode(PNameCoder, Object, java.lang.Class, JormFactory)
     */
    Object encode(PName pn) throws PException;
    
    /**
     * clean all naming Manager 
     *
     */
    void clean();
}
