/**
 * Speedo: an implementation of JDO compliant personality on top of JORM generic
 * I/O sub-system.
 * Copyright (C) 2001-2006 France Telecom
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: speedo@objectweb.org
 *
 * Authors: S. Chassande-Barrioz. A. Lefebvre
 *
 */

package org.objectweb.speedo.ejb.runtime.query.parser;

import junit.framework.TestCase;

import java.io.CharArrayReader;

import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.Monolog;
import org.objectweb.speedo.query.ejb.parser.EJBQL;

/**
 * Speedo QL Test
 * Test a list of valid filters.
 */
public class EJBQLTest extends TestCase {
    final static String filters[] = {
        //tests from specification
        "SELECT DISTINCT o FROM Order AS o JOIN o.lineItems AS l WHERE l.shipped = FALSE",
        "SELECT DISTINCT o FROM Order o JOIN o.lineItems l JOIN l.product p " 
        		+ "WHERE p.productType = 'office_supplies'",
        "SELECT DISTINCT o1 FROM Order o1, Order o2 " 
        		+ "WHERE o1.quantity > o2.quantity AND " 
        		+ "o2.customer.lastname = 'Smith' AND " 
        		+ "o2.customer.firstname= 'John'",
        "SELECT DISTINCT l.product FROM Order AS o, IN(o.lineItems) l",
        "select c from Customer c, Employee e where c.hatsize = e.shoesize",
        "SELECT c FROM Customer c JOIN c.orders o WHERE c.status = 1",
        "SELECT c FROM Customer c INNER JOIN c.orders o WHERE c.status = 1",
        "SELECT OBJECT(c) FROM Customer c, IN(c.orders) o WHERE c.status = 1",
        "SELECT c FROM Customer c LEFT JOIN c.orders o WHERE c.status = 1",
        "SELECT c FROM Customer c LEFT OUTER JOIN c.orders o WHERE c.status = 1",
        "SELECT d FROM Department d LEFT JOIN FETCH d.employees WHERE d.deptno = 1",
        "SELECT DISTINCT o FROM Order o JOIN o.lineItems l JOIN l.product p " 
        		+ "WHERE p.productType = 'office_supplies'",
        "SELECT DISTINCT o FROM Order o, IN(o.lineItems) l " 
        		+ "WHERE l.product.productType = 'office_supplies'",
        "SELECT o FROM Order AS o, IN(o.lineItems) l, Product p",
        //TODO handled names parameters
        "SELECT c FROM Customer c WHERE c.status = :stat",
        "SELECT o FROM Order o WHERE o.x BETWEEN o.y AND o.z",
        "SELECT o FROM Order o WHERE o.y <= o.x AND o.x <= o.z",
        "SELECT p FROM Order p WHERE p.age BETWEEN 15 and 19",
        "SELECT p FROM Order p WHERE p.age >= 15 AND p.age <= 19",
        "SELECT p FROM Order p WHERE p.age NOT BETWEEN 15 and 19",
        "SELECT p FROM Order p WHERE p.age < 15 OR p.age > 19",
        "SELECT o FROM Order o WHERE o.country IN ('UK','US','France')",
        "SELECT o FROM Order o WHERE (o.country='UK') OR (o.country='US') OR (o.country='France')",
        "SELECT DISTINCT emp FROM Employee emp WHERE EXISTS ("
			+ "SELECT spouseEmp FROM Employee spouseEmp WHERE spouseEmp = emp.spouse)",
        "SELECT c FROM Customer c WHERE (SELECT COUNT(o) FROM c.orders o) > 10",
		"SELECT goodCustomer FROM Customer goodCustomer "
        		+ "WHERE goodCustomer.balanceOwed < "
        		+ "(SELECT avg(c.balanceOwed) FROM Customer c)",
        "SELECT c.status, avg(c.filledOrderCount), count(c) FROM Customer c "
        		+ "GROUP BY c.status HAVING c.status IN (1, 2)",
        "SELECT c.country, COUNT(c) FROM Customer c GROUP BY c.country " 
        		+ "HAVING COUNT(c.country) > 3",
        //TODO: handle any keyword as part of a path - replacing with ccount for now
        //	"SELECT c.id, c.status FROM Customer c JOIN c.orders o WHERE o.count > 100",
        "SELECT c.id, c.status FROM Customer c JOIN c.orders o WHERE o.ccount > 100",
        //TODO: handle any keyword as part of a path - replacing with ccount for now
        //"SELECT NEW com.acme.example.CustomerDetails(c.id, c.status, o.count) "
        	//	+ "FROM Customer c JOIN c.orders o WHERE o.count > 100",
        "SELECT NEW com.acme.example.CustomerDetails(c.id, c.status, o.ccount) "
        		+ "FROM Customer c JOIN c.orders o WHERE o.ccount > 100",
        "SELECT AVG(o.quantity) FROM Order o",
        "SELECT SUM(l.price) FROM Order o JOIN o.lineItems l JOIN o.customer c " 
        		+ "WHERE c.lastname = 'Smith' AND c.firstname = 'John'",
        "SELECT COUNT(o) FROM Order o",
        "SELECT COUNT(l.price) FROM Order o JOIN o.lineItems l JOIN o.customer c "
        		+ "WHERE c.lastname = 'Smith' AND c.firstname = 'John'",
        "SELECT COUNT(l) FROM Order o JOIN o.lineItems l JOIN o.customer c "
        		+ "WHERE c.lastname = 'Smith' AND c.firstname = 'John' "
        		+ "AND l.price IS NOT NULL",
        "SELECT o FROM Customer c JOIN c.orders o JOIN c.address a "
        		+ "WHERE a.state = 'CA' ORDER BY o.quantity, o.totalcost",
        "SELECT o.quantity, a.zipcode FROM Customer c JOIN c.orders o JOIN c.address a "
        		+ "WHERE a.state = 'CA' ORDER BY o.quantity, a.zipcode",
        	//the following one is identified as not legal in the spec because orderby
        	// is not reflected in the select clause
        		/*
        	"SELECT p.product_name FROM Order o JOIN o.lineItems l JOIN l.product p JOIN o.customer c "
        		+ "WHERE c.lastname = 'Smith' AND c.firstname = 'John' "
        		+ "ORDER BY p.price", 
        */
        "DELETE FROM Customer c WHERE c.status = 'inactive'", 
        "DELETE FROM Customer c WHERE c.status = 'inactive' AND c.orders IS EMPTY", 
        "UPDATE customer c SET c.status = 'outstanding' "
        		+ "WHERE c.balance < 10000  AND 1000 > "
        		+ "(SELECT COUNT(o) FROM customer cust JOIN cust.order o)",
  		//tests from specification: Examples
        "SELECT o FROM Order o",
        "SELECT o FROM Order o WHERE o.shippingAddress.state = 'CA'",        
        "SELECT DISTINCT o.shippingAddress.state FROM Order o",
        "SELECT o FROM Order o WHERE o.lineItems IS NOT EMPTY",
        "SELECT o FROM Order o WHERE o.lineItems IS EMPTY",
        "SELECT DISTINCT o FROM Order o JOIN o.lineItems l WHERE l.shipped = FALSE",
        "SELECT o FROM Order o WHERE "
         	+ "NOT (o.shippingAddress.state = o.billingAddress.state "
         	+ "AND o.shippingAddress.city = o.billingAddress.city "
         	+ "AND o.shippingAddress.street = o.billingAddress.street)",
         "SELECT o FROM Order o WHERE o.shippingAddress <> o.billingAddress",
         "SELECT DISTINCT o FROM Order o JOIN o.lineItems l " 
         	+ "WHERE l.product.type = 'book' AND " 
         	+ "l.product.name = 'Applying Enterprise JavaBeans: Component-Based Development for the J2EE Platform'", 
         "SELECT DISTINCT o FROM Order o, IN(o.lineItems) l WHERE l.product.name = ?1",
         //additional tests
         "SELECT o.shippingAddress.state FROM Order o",
         "SELECT o FROM Order o WHERE o.shippingAddress.state IS NOT EMPTY",
         "SELECT o FROM Order o WHERE o.shippingAddress.state = ?1",
         "SELECT o FROM Order o WHERE o.shippingAddress.state = o.state",
         "SELECT o FROM Order o WHERE 'CA' = o.shippingAddress.state",
         "SELECT o FROM Order o WHERE 12 = o.shippingAddress.state",
         "SELECT o FROM Order o WHERE o.shippingAddress.state = 12"

    };

    private final static String LOGGER_NAME
            = "org.objectweb.speedo.rt.query.parser.ejbql";

    private static Logger logger = null;

    static {
        logger = Monolog.initialize().getLogger(LOGGER_NAME);
    }


    public EJBQLTest(String name) {
        super(name);
    }

    public void testAll() {
        logger.log(BasicLevel.DEBUG, "EJBQL Test - javacc grammar");
        for (int i = 0; i < filters.length; i++) {
            String filter = filters[i];
            logger.log(BasicLevel.DEBUG, "Testing: " + filter);
            try {
                new EJBQL(new CharArrayReader(filter.toCharArray())).EJBQL();
            } catch (Throwable e) {
                logger.log(BasicLevel.ERROR, "Parsing of: " + filter, e);
                fail("Parsing of: " + filter);
            }

        }
    }
}
