/*
 * Speedo: an implementation of JDO compliant personality on top of JORM
 * generic I/O sub-system. Copyright (C) 2001-2004 France Telecom R&D
 * 
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * Release: 1.0
 * 
 * Created on Apr 8, 2004 @author fmillevi@yahoo.com
 *  
 */
package org.objectweb.speedo.j2eedo.bo;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Vector;

import javax.transaction.Synchronization;

/**
 * The J2eedo application maintains three static lists storing the primary key
 * of existing
 * {@link org.objectweb.speedo.j2eedo.database.Department departments},
 * {@link org.objectweb.speedo.j2eedo.database.Employee employees}and
 * {@link org.objectweb.speedo.j2eedo.database.Project projects}. This class
 * <ul>
 * <li>Each time a new instance of this object is created the factory class
 * has to add the id in the dedicated poll</li>
 * <li>Each time an existing object is deleted, its id have to be remove from
 * the poll</li>
 * </ul>
 * <p>
 * This class allows the <b>synchronization</b> the <b>static pools</b> id
 * content update and the <b>validation of the transaction</b>.
 * 
 * @author fmillevi@yahoo.com
 * @see DatabaseImpl
 */
public class PollsSynchronizations implements Synchronization {
	private ArrayList pools;
	private ArrayList actionADDLists;
	private ArrayList actionREMLists;
	private boolean processed;

    /**
     * The constructor initialize the list of add and remove actions to be
     * treated while performing the {@link #afterCompletion afterCompletion}
     * 
     * @see org.objectweb.speedo.j2eedo.bo.DatabaseImpl#poolOfDepartmentId
     * @see org.objectweb.speedo.j2eedo.bo.DatabaseImpl#poolOfEmployeeId
     * @see org.objectweb.speedo.j2eedo.bo.DatabaseImpl#poolOfProjectId
     */
    public PollsSynchronizations() {
        this.pools = new ArrayList();
        this.actionADDLists = new ArrayList();
        this.actionREMLists = new ArrayList();
        this.processed = false;
    }

    /**
	 * Do nothing
	 * 
	 * @see javax.transaction.Synchronization#beforeCompletion()
	 */
	public void beforeCompletion() {
	}

	/**
	 * The afterCompletion method has to update the pool content according the
	 * action list.
	 * 
	 * @see javax.transaction.Synchronization#afterCompletion(int)
	 */
	public void afterCompletion(int arg0) {
		if (processed) 
			return;
		processed = true;
		for (int i = 0; i < pools.size(); i++) {
            Vector pool = (Vector) pools.get(i);
            ArrayList v = (ArrayList) actionADDLists.get(i); 
            for (int j = 0; j < v.size(); j++) {
                pool.add(v.get(j));
            }
            v = (ArrayList) actionREMLists.get(i); 
            for (int j = 0; j < v.size(); j++) {
				pool.remove(v.get(j));
			}
		}
	}

	/**
	 * Remember the new id created
	 * 
	 * @param id
	 */
	public void addInPool(Vector pool, long id) {
        Collection[] ps = getPoolIndex(pool);
        Long lid = new Long(id);
        if (!ps[1].contains(lid)) {
            ps[1].add(lid);
        }
	}

	/**
	 * Remember the new id deleted
	 * 
	 * @param id
	 */
	public void removeFromPool(Vector pool, long id) {
        Collection[] ps = getPoolIndex(pool);
        Long lid = new Long(id);
		if (!ps[2].contains(lid)) {
            ps[2].add(lid);
        }
	}
    
    private Collection[] getPoolIndex(Vector pool) {
        int i = 0;
        while(i< pools.size() && pool != pools.get(i)) {
            i++;
        }
        if (i == pools.size()) {
            pools.add(pool);
            actionADDLists.add(new ArrayList());
            actionREMLists.add(new ArrayList());
        }
        return new Collection[]{
                (Collection) pools.get(i),
                (Collection) actionADDLists.get(i),
                (Collection) actionREMLists.get(i)
        };
    }

}
