/**
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.stress;

import org.objectweb.speedo.pobjects.userid.LinkedIntUserId;
import org.objectweb.util.monolog.api.BasicLevel;

import junit.framework.Assert;

import javax.jdo.PersistenceManager;
import javax.jdo.JDOFatalException;

/**
 *
 * @author S.Chassande-Barrioz
 */
public abstract class LinkedIntUserIdHelper extends StressHelper {

	/**
	 * is the lists of object identifier prepared before the transaction
	 * execution.
	 * if (oids == null) {
	 *   db is not initialised
	 * } else if (oids != null && oids.length==0) {
	 *   db initialised and keepOid = false
	 * } else {
	 *   db initialised and keepOid == true
	 * }
	 */
	protected static Object[] oids = null;
	protected String DBSIZE = getLoggerName() + ".dbsize";
	protected String NO_DB_INIT = getLoggerName() + ".nodbinit";


	public LinkedIntUserIdHelper(String s) {
		super(s);
	}

	protected String[] getClassNamesToInit() {
		return new String[]{LinkedIntUserId.class.getName()};
	}

	protected boolean linkObject() {
		return true;
	}

	protected boolean keepOid() {
		return true;
	}

	/**
	 * IMPORTANT: dot not removed data on support in order to avoid next
	 * creations
	 */
	public void setUp() throws Exception {
		logger.log(BasicLevel.DEBUG, "LinkedIntUserIdHelper.setUp");
		cleanup();
		initDataStructure(false);
	}

	/**
	 * Creates the persistent object if it is not already done.
	 * @param task the task to prepare
	 * @param _ctx the context of the test.
	 */
	protected void prepareTask(Task task, Object _ctx) {
		debug = logger.isLoggable(BasicLevel.DEBUG);
		logger.log(BasicLevel.INFO, "LinkedIntUserIdHelper.debug="+debug);
		super.prepareTask(task, _ctx);
		LIUICtx ctx = (LIUICtx) _ctx;
		if (oids == null) {
			synchronized (TestGetObject.class) {
				if (oids == null && !Boolean.getBoolean(NO_DB_INIT)) {
					//Initialisation the database
					logger.log(BasicLevel.INFO, "\tPreparing test...");
					new PrepareTestLinked(this)
					        .prepare(ctx, keepOid(), linkObject());
					if (keepOid()) {
						//keep oids in the static variable
						oids = ctx.oids;
					} else {
						//db initialized without oids
						oids = new Object[0];
					}
					logger.log(BasicLevel.INFO, "\tTest Prepared.");
				}
			}
		}
		ctx.oids = oids;
		
	}

	
	/**
	 * The context to use for the LinkedIntUserId object
	 */
	public class LIUICtx {
		/**
		 * The identifier of the created object (see keepOid method)
		 */
		public Object oids[];
		/**
		 * The number of persistent object
		 */
		public int dbSize;
		/**
		 * indicates if the oid must be kept during the preparation of the task
		 */
		public boolean keepOidOnPrepare;
		public boolean linkOnPrepare;
		public boolean endPrepare;
		public Object[] currentsOnPrepare;
		public Object[] nextsOnPrepare;

		public LIUICtx(int dbSize) {
			this.dbSize = dbSize;
			oids = new Object[dbSize];
		}

		public void initOnPrepare(int nbTx, boolean fetchOid, boolean linkOnPrepare) {
			this.keepOidOnPrepare = fetchOid;
			this.linkOnPrepare = linkOnPrepare;
			this.endPrepare = false;
			nextsOnPrepare = new Object[nbTx];
			currentsOnPrepare = new Object[nbTx];
		}

		public void endOnPrepare(int nbTx, boolean fetchOid, boolean linkOnPrepare) {
			this.endPrepare = true;
		}

		public String toString() {
			return "dbSize = " + dbSize;
		}		
	}
}

class PrepareTestLinked extends StressHelper {

	private final static int NB_CREATION = 1000;
	private final static int NB_THREAD = 4;

	public PrepareTestLinked(LinkedIntUserIdHelper helper) {
		super(helper.getName());
	}

	protected String[] getClassNamesToInit() {
		return new String[]{LinkedIntUserId.class.getName()};
	}

	protected String getLoggerName() {
		return STRESS_LOG_NAME + ".LinkedIntUserIdHelper";
	}

	protected String getLogPrefix() {
		return super.getLogPrefix() + "\t";
	}

	protected void perform(StressHelper.Task task,
	                       int threadId,
	                       int txId,
	                       Object ctx,
	                       PerformResult res) {
		LinkedIntUserIdHelper.LIUICtx pctx = (LinkedIntUserIdHelper.LIUICtx) ctx;
		int plus = pctx.dbSize % NB_CREATION;
		int nbTx = task.txToExecute.length;
		PersistenceManager pm = getPM(task, threadId, txId);
		try {
			res.beginTest();
			beginTx(pm, task, threadId, txId);
			LinkedIntUserId current=null, next=null;
			
			if (!pctx.linkOnPrepare) {
				if (plus > 0) {
					//The first transaction creates the additional object
					for (int oid = 0; txId == 0 && oid < plus; oid++) {
						current = new LinkedIntUserId(oid, "Obj No " + oid, null);
						pm.makePersistent(current);
						if (pctx.keepOidOnPrepare) {
							pctx.oids[oid] = pm.getObjectId(current);
						}
					}						
				} 
				// The other transactions create 'nbCreation' objects
				for (int no = 0; no < NB_CREATION; no++) {
					int oid = (txId * NB_CREATION) + no + plus;
					if (oid < pctx.dbSize) {
						current = new LinkedIntUserId(oid, "Obj No " + oid, null);
						pm.makePersistent(current);
						if (pctx.keepOidOnPrepare) {
							pctx.oids[oid] = pm.getObjectId(current);
						}
					}
				}
			} else if (!pctx.endPrepare) {
					
				/*
				 * First : save the oid of objects to be linked between transactions   
				 */				
				
				int oid = 0;
				if (plus > 0 && txId == 0) {
					current = new LinkedIntUserId(oid, "Obj No " + oid, null);
					pm.makePersistent(current);
					if (pctx.keepOidOnPrepare) {
						pctx.oids[oid] = pm.getObjectId(current);
					}
					
					oid++;
					next = new LinkedIntUserId(oid, "Obj No " + oid, null);
					pm.makePersistent(next);
					if (pctx.keepOidOnPrepare) {
						pctx.oids[oid] = pm.getObjectId(next);
					}
					
					current.setNext(next);
					pctx.currentsOnPrepare[txId] = pm.getObjectId(current);
					
					//The first transaction creates the additional object
					for (oid = 2; oid < plus; oid++) {				
						current = new LinkedIntUserId(oid, "Obj No " + oid, null);
						pm.makePersistent(current);
						if (pctx.keepOidOnPrepare) {
							pctx.oids[oid] = pm.getObjectId(current);
						}
						next.setNext(current);
						pm.makePersistent(next);
						next = current;						
					}				
				} 
				
				// The other transactions create 'nbCreation' objects
				for (int no = 0; no < NB_CREATION; no++) {
					oid = (txId * NB_CREATION) + no + plus;
					if (oid < pctx.dbSize) {					
						current = new LinkedIntUserId(oid, "Obj No " + oid, null);
						pm.makePersistent(current);
						if (no == 0 && pctx.currentsOnPrepare[txId] == null) {
							pctx.currentsOnPrepare[txId] = pm.getObjectId(current);
							
						}
						if (pctx.keepOidOnPrepare) {
							pctx.oids[oid] = pm.getObjectId(current);
						}
						if (next != null) {
							next.setNext(current);
							pm.makePersistent(next);
						}
						next = current;
					}
				}
				
				// save the oid of objects to be linked between transactions
				pctx.nextsOnPrepare[txId] = pm.getObjectId(next);				
				
			} else if (txId == nbTx - 1){ // do it once only
			
				/*
				 * Second : link the oid    
				 */				
				if (debug) {
					for (int i=0; i < nbTx; i++) {						
						logger.log(BasicLevel.DEBUG, "currentsOnPrepare["+i+"]="+ pctx.currentsOnPrepare[i]);
						logger.log(BasicLevel.DEBUG, "nextsOnPrepare["+i+"]="+ pctx.nextsOnPrepare[i]);					
					}
				}
				
				// make circular list				
				next = (LinkedIntUserId) pm.getObjectById(pctx.currentsOnPrepare[0], false);
				current = (LinkedIntUserId) pm.getObjectById(pctx.nextsOnPrepare[txId], false);
				
				if (plus != 0) {									
					if (txId == 0) {
						current = (LinkedIntUserId) pm.getObjectById(pctx.nextsOnPrepare[txId], false);
					} else {
						current = (LinkedIntUserId) pm.getObjectById(pctx.nextsOnPrepare[txId-1], false);
					}
				}
				
				current.setNext(next);
				pm.makePersistent(current);						
				if (pctx.keepOidOnPrepare) {
					int indice = (nbTx * NB_CREATION) - 1;
					if (plus != 0) {			
						indice = (txId * NB_CREATION) + plus - 1;
					}					
					pctx.oids[indice] = pm.getObjectId(current);
				}
				
				
				// link objects between transactions
				for (int no = 0; no < nbTx - 1; no++) {												
					if (pctx.currentsOnPrepare[no+1] != null) {
						next = (LinkedIntUserId) pm.getObjectById(pctx.currentsOnPrepare[no+1], false);					
						current = (LinkedIntUserId) pm.getObjectById(pctx.nextsOnPrepare[no], false);					
						current.setNext(next);						
					}
				}
				
				// verify all
				for (int i=0; i<pctx.oids.length-1; i++) {
					Assert.assertNotNull("i="+i, pctx.oids[i]);			
					Assert.assertNotNull("i="+i+1, pctx.oids[i+1]);
					current = (LinkedIntUserId)  pm.getObjectById(pctx.oids[i], false);
					next = (LinkedIntUserId)  pm.getObjectById(pctx.oids[i+1], false);
					Assert.assertEquals("name="+current.getName(), current.getNext().getName(), next.getName());
				}
				next = (LinkedIntUserId)  pm.getObjectById(pctx.oids[0], false);
				current = (LinkedIntUserId)  pm.getObjectById(pctx.oids[pctx.oids.length-1], false);
				Assert.assertEquals("name="+current.getName(), current.getNext().getName(), next.getName());
				
				
			}
			commitTx(pm, task, threadId, txId);
			res.endTest();
		} catch (JDOFatalException e) {
			rollbackOnException(pm, e, res, task, threadId, txId);
		} catch (Throwable e) {
			stopOnError(pm, e, res, task, threadId, txId);
		} finally {
			closePM(pm, threadId, txId, task, res);
		}

	}

	public void prepare(LinkedIntUserIdHelper.LIUICtx ctx, boolean fetchOid, boolean link) {
		int plus = ctx.dbSize % NB_CREATION;
		int nbTx = (ctx.dbSize / NB_CREATION) + (plus > 0 ? 1 : 0);
		ctx.initOnPrepare(nbTx, fetchOid, link);
		perform(NB_THREAD, nbTx, Integer.getInteger(TIMEOUT, 200000).intValue(), ctx);
		ctx.endOnPrepare(nbTx, fetchOid, link);
		perform(NB_THREAD, nbTx, Integer.getInteger(TIMEOUT, 200000).intValue(), ctx);
		
	}
}

