/**
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package org.objectweb.speedo.stress;

import java.util.ArrayList;
import java.util.List;

import javax.jdo.JDOFatalException;
import javax.jdo.PersistenceManager;

import junit.framework.Assert;

import org.objectweb.speedo.pobjects.collection.AllCollection;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * Stress the collection access function of Speedo.
 *
 * @author S.Rodiere
 */
public class TestCollection extends StressHelper {
	static private int nbBeginTest = 0;
	
	public TestCollection(String s) {
		super(s);
	}

	protected String[] getClassNamesToInit() {
		return new String[]{AllCollection.class.getName()};
	}

	protected String getLoggerName() {
		return STRESS_LOG_NAME + ".TestCollection";
	}

	protected void perform(Task task, int threadId, int txId, Object ctx, PerformResult res) {
		synchronized (this){
			nbBeginTest++;
		}	
		long oids = (nbBeginTest * 6);
		PersistenceManager pm = getPM(task, threadId, txId);
		try {
			res.beginTest();
			// Execute first transaction on persistent Collections
			beginTx(pm, task, threadId, txId);
			long[] ls = new long[]{oids, oids + 1, oids + 2, oids + 3, oids + 4, oids + 5};
			long[] reverse = new long[]{oids + 5, oids + 4, oids + 3, oids + 2, oids + 1, oids};
			AllCollection ac = new AllCollection("stressCollection" + nbBeginTest);
			ac.setLongs(ls);
			ac.setRefs(new Object[]{ac});
			pm.makePersistent(ac);
			commitTx(pm, task, threadId, txId);

			// Reverse the data
			ac.setLongs(reverse);
			ac.setRefs(new Object[]{ac});

			// Verify result
			List expectedLong = new ArrayList(reverse.length);
			for (int i = 0; i < reverse.length; i++) {
				expectedLong.add(new Long(reverse[i]));
			}

			List expectedRef = new ArrayList(1);
			expectedRef.add(ac);

			//Collection
			Assert.assertNotNull("The collection of Long is null", ac.getCol_Long());
			assertSameCollection("The collection of Long", expectedLong, ac.getCol_Long());
			Assert.assertNotNull("The collection of reference is null", ac.getCol_ref());
			assertSameCollection("The collection of ref", expectedRef, ac.getCol_ref());

			//HashSet
			Assert.assertNotNull("The Hashset of Long is null", ac.getHset_Long());
			assertSameCollection("The Hashset of Long", expectedLong, ac.getHset_Long());
			Assert.assertNotNull("The Hashset of reference is null", ac.getHset_ref());
			assertSameCollection("The Hashset of ref", expectedRef, ac.getHset_ref());

			//List
			Assert.assertNotNull("The list of Long is null", ac.getList_Long());
			assertSameList("The list of Long", expectedLong, ac.getList_Long());
			Assert.assertNotNull("The list of reference is null", ac.getList_ref());
			assertSameList("The list of reference", expectedRef, ac.getList_ref());

			//Set
			Assert.assertNotNull("The set of Long is null", ac.getSet_Long());
			assertSameCollection("The set of Long", expectedLong, ac.getSet_Long());
			Assert.assertNotNull("The set of reference is null", ac.getSet_ref());
			assertSameCollection("The set of ref", expectedRef, ac.getSet_ref());

			// Execute second transaction on persistent Collections
			beginTx(pm, task, threadId, txId);
			logger.log(BasicLevel.DEBUG, ac.getId() + " is being deleted");
			pm.deletePersistent(ac);
			commitTx(pm, task, threadId, txId);
			res.endTest();
			
		} catch (JDOFatalException e) {
			rollbackOnException(pm, e, res, task, threadId, txId);
		} catch (Throwable e) {
			stopOnError(pm, e, res, task, threadId, txId);
		} finally {
			closePM(pm, threadId, txId, task, res);
		}
	}

	/**
	 * Tests the read of a lot of persistent objects, with interactive setting
	 * of test parameteres (see file userconf/project.properties).
	 */
	public void testCollection() {
		if (interactive) {
			perform(Integer.getInteger(THREAD, 1).intValue(),
			        Integer.getInteger(TX, 9).intValue(),
			        Integer.getInteger(TIMEOUT, 200000).intValue(), null);
		}
	}

	/**
	 * Tests 100 transactions which stress collections using 1 thread.
	 */
	public void testCollectionTh1Tx100() {
		if (!interactive) {
			logger.log(BasicLevel.INFO, "testCollectionTh1Tx100");
			perform(1, 100, 1000000, null);
		}
	}

	/**
	 * Tests 1.000 transactions which stress collections using 1 thread.
	 */
	public void testCollectionTh1Tx1000() {

		if (!interactive) {
			logger.log(BasicLevel.INFO, "testCollectionTh1Tx1000");
			perform(1, 1000, 1000000, null);
		}
	}

	/**
	 * Tests 10.000 transactions which stress collections using 1 thread.
	 */
	public void testCollectionTh1Tx10000() {

		if (!interactive) {
			logger.log(BasicLevel.INFO, "testCollectionTh1Tx10000");
			perform(1, 10000, 1000000, null);
		}
	}

	/**
	 * Tests 100.000 transactions which stress collections using 1 thread.
	 */
	public void _testCollectionTh1Tx100000() {

		if (!interactive) {
			logger.log(BasicLevel.INFO, "testCollectionTh1Tx100000");
			perform(1, 100000, 1000000, null);
		}
	}

	/**
	 * Tests 100 transactions which stress collections using 2 thread.
	 */
	public void testCollectionTh2Tx100() {

		if (!interactive) {
			logger.log(BasicLevel.INFO, "testCollectionTh2Tx100");
			perform(2, 100, 1000000, null);
		}
	}

	/**
	 * Tests 1000 transactions which stress collections using 2 thread.
	 */
	public void testCollectionTh2Tx1000() {

		if (!interactive) {
			logger.log(BasicLevel.INFO, "testCollectionTh2Tx1000");
			perform(2, 1000, 1000000, null);
		}
	}

	/**
	 * Tests 10.000 transactions which stress collections using 2 thread.
	 */
	public void testCollectionTh2Tx10000() {

		if (!interactive) {
			logger.log(BasicLevel.INFO, "testCollectionTh2Tx10000");
			perform(2, 10000, 1000000, null);
		}
	}

	/**
	 * Tests 100.000 transactions which stress collections using 2 thread.
	 */
	public void _testCollectionTh2Tx100000() {

		if (!interactive) {
			logger.log(BasicLevel.INFO, "testCollectionTh2Tx100000");
			perform(2, 100000, 1000000, null);
		}
	}


	/**
	 * Tests 1000 transactions which stress collections using 10 thread.
	 */
	public void testCollectionTh10Tx1000() {

		if (!interactive) {
			logger.log(BasicLevel.INFO, "testCollectionTh10Tx1000");
			perform(10, 1000, 1000000, null);
		}
	}

	/**
	 * Tests 10.000 transactions which stress collections using 10 thread.
	 */
	public void testCollectionTh10Tx10000() {

		if (!interactive) {
			logger.log(BasicLevel.INFO, "testCollectionTh10Tx10000");
			perform(10, 10000, 1000000, null);
		}
	}

	/**
	 * Tests 100.000 transactions which stress collections using 10 thread.
	 */
	public void _testCollectionTh10Tx100000() {

		if (!interactive) {
			logger.log(BasicLevel.INFO, "testCollectionTh10Tx100000");
			perform(10, 100000, 1000000, null);
		}
	}

	/**
	 * Tests 100 transactions which stress collections using 100 thread.
	 */
	public void testCollectionTh100Tx100() {

		if (!interactive) {
			logger.log(BasicLevel.INFO, "testCollectionTh100Tx100");
			perform(100, 100, 1000000, null);
		}
	}

	/**
	 * Tests 1000 transactions which stress collections using 1000 thread.
	 */
	public void testCollectionTh1000Tx1000() {
		
		if (!interactive) {
			logger.log(BasicLevel.INFO, "testCollectionTh1000Tx1000");
			perform(1000, 1000, 1000000, null);
		}
	}

}
