/**
 * Copyright (C) 2001-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package org.objectweb.speedo.stress;

import org.objectweb.speedo.pobjects.relations.C;
import org.objectweb.speedo.pobjects.relations.D;
import org.objectweb.util.monolog.api.BasicLevel;

import javax.jdo.JDOFatalException;
import javax.jdo.PersistenceManager;

/**
 * Stress the relation one-one implementation by Speedo.
 *
 * @author M. Guillemin
 */
public class TestRelation1_1Concurrency extends CDRelationHelper {

	public TestRelation1_1Concurrency(String s) {
		super(s);
	}

	protected String[] getClassNamesToInit() {
		return new String[]{
			org.objectweb.speedo.pobjects.relations.C.class.getName(),
			org.objectweb.speedo.pobjects.relations.D.class.getName()
		};
	}

	protected String getLoggerName() {
		return STRESS_LOG_NAME + ".TestRelation1_1Concurrency";
	}

	protected int getNbSucc() {
		return 2; // because d1, d2
	}
	
	protected void perform(StressHelper.Task task,
	                       int threadId,
	                       int txId,
	                       Object ctx,
	                       StressHelper.PerformResult res) {
		CDCtx cdctx = (CDCtx) ctx;
		PersistenceManager pm = getPM(task, threadId, txId);
		try {
			res.beginTest();			
			beginTx(pm, task, threadId, txId);
			
			int oid = getOid(cdctx, true, getNbSucc());

			C c1 = (C) pm.getObjectById(cdctx.oidsC[oid], false);
			C c2 = (C) pm.getObjectById(cdctx.oidsC[oid+1], false);
			D d1 = (D) pm.getObjectById(cdctx.oidsD[oid], false);
			D d2 = (D) pm.getObjectById(cdctx.oidsD[oid+1], false);

			c1.setD(d1);
			assertTrue(d1.getC() == c1);
			c2.setD(d2);
			assertTrue(d2.getC() == c2);
			c1.setD(d2);
			assertTrue(d1.getC() == null);
			assertTrue(d2.getC() == c1);
			assertTrue(c2.getD() == null);
			d1.setC(c2);
			assertTrue(c2.getD() == d1);
			d1.setC(c1);
			assertTrue(c1.getD() == d1);
			assertTrue(d2.getC() == null);
			assertTrue(c2.getD() == null);

			commitTx(pm, task, threadId, txId);

			res.endTest();
		} catch (JDOFatalException e) {
			rollbackOnException(pm, e, res, task, threadId, txId);
		} catch (Throwable e) {
			stopOnError(pm, e, res, task, threadId, txId);
		} finally {
			closePM(pm, threadId, txId, task, res);
		}
	}

	/**
	 * Tests the read of a lot of persistent objects, with interactive setting
	 * of test parameteres (see file userconf/project.properties).
	 */
	public void testIntercative() {
		if (interactive) {
			logger.log(BasicLevel.INFO, "testInteractive");
			perform(Integer.getInteger(THREAD, 1).intValue(),
			        Integer.getInteger(TX, 1).intValue(),
			        Integer.getInteger(TIMEOUT, 1).intValue(),
					new CDCtx(Integer.getInteger(DBSIZE, 1000).intValue()));
		} else {
			logger.log(BasicLevel.INFO, "Non Interactive mode: ignore testInteractive");
		}				
	}

	/**
	 * Methode d'aiguillage
	 */
	private void perform(
			int nbThread,
			int dbSize,
			int nbTx,
			int threadTimeout) {
		perform(nbThread, nbTx, threadTimeout, new CDCtx(dbSize));
	}

	/**
	 * Tests 100 transactions where each of them stress 1_1 linked persistents objects using 1 thread.
	 */
	public void testRelation1_1ConcurrencyTh1Tx100() {
		if (!interactive) {
			logger.log(BasicLevel.INFO, "testRelation1_1ConcurrencyTh1Tx100");
			perform(1, 10000, 100, 1000000);
		} else {
			logger.log(BasicLevel.INFO, "Non Interactive mode: ignore testRelation1_1ConcurrencyTh1Tx100");
		}
	}

	/**
	 * Tests 1000 transactions where each of them stress 1_1 linked persistents objects using 1 thread.
	 */
	public void testRelation1_1ConcurrencyTh1Tx1000() {
		if (!interactive) {
			logger.log(BasicLevel.INFO, "testRelation1_1ConcurrencyTh1Tx1000");
			perform(1, 10000, 1000, 1000000);
		} else {
			logger.log(BasicLevel.INFO, "Non Interactive mode: ignore testRelation1_1ConcurrencyTh1Tx1000");
		}
	}

	/**
	 * Tests 10000 transactions where each of them stress 1_1 linked persistents objects using 1 thread.
	 */
	public void testRelation1_1ConcurrencyTh1Tx10000() {
		if (!interactive) {
			logger.log(BasicLevel.INFO, "testRelation1_1ConcurrencyTh1Tx10000");
			perform(1, 10000, 10000, 1000000);
		} else {
			logger.log(BasicLevel.INFO, "Non Interactive mode: ignore testRelation1_1ConcurrencyTh1Tx10000");
		}
	}

	/**
	 * Tests 100 transactions where each of them stress 1_1 linked persistents objects using 2 thread.
	 */
	public void testRelation1_1ConcurrencyTh2Tx100() {
		if (!interactive) {
			logger.log(BasicLevel.INFO, "testRelation1_1ConcurrencyTh2Tx100");
			perform(2, 10000, 100, 1000000);
		} else {
			logger.log(BasicLevel.INFO, "Non Interactive mode: ignore testRelation1_1ConcurrencyTh2Tx100");
		}
	}

	/**
	 * Tests 1000 transactions where each of them stress 1_1 linked persistents objects using 2 thread.
	 */
	public void testRelation1_1ConcurrencyTh2Tx1000() {
		if (!interactive) {
			logger.log(BasicLevel.INFO, "testRelation1_1ConcurrencyTh2Tx1000");
			perform(2, 10000, 1000, 1000000);
		} else {
			logger.log(BasicLevel.INFO, "Non Interactive mode: ignore testRelation1_1ConcurrencyTh2Tx1000");
		}
	}

	/**
	 * Tests 10000 transactions where each of them stress 1_1 linked persistents objects using 2 thread.
	 */
	public void testRelation1_1ConcurrencyTh2Tx10000() {
		if (!interactive) {
			logger.log(BasicLevel.INFO, "testRelation1_1ConcurrencyTh2Tx10000");
			perform(2, 10000, 10000, 1000000);
		} else {
			logger.log(BasicLevel.INFO, "Non Interactive mode: ignore testRelation1_1ConcurrencyTh2Tx10000");
		}
	}

	/**
	 * Tests 100 transactions where each of them stress 1_1 linked persistents objects using 10 thread.
	 */
	public void testRelation1_1ConcurrencyTh10Tx100() {
		if (!interactive) {
			logger.log(BasicLevel.INFO, "testRelation1_1ConcurrencyTh10Tx100");
			perform(10, 10000, 100, 1000000);
		} else {
			logger.log(BasicLevel.INFO, "Non Interactive mode: ignore testRelation1_1ConcurrencyTh10Tx100");
		}
	}

	/**
	 * Tests 1000 transactions where each of them stress 1_1 linked persistents objects using 10 thread.
	 */
	public void testRelation1_1ConcurrencyTh10Tx1000() {
		if (!interactive) {
			logger.log(BasicLevel.INFO, "testRelation1_1ConcurrencyTh10Tx1000");
			perform(10, 10000, 1000, 1000000);
		} else {
			logger.log(BasicLevel.INFO, "Non Interactive mode: ignore testRelation1_1ConcurrencyTh10Tx1000");
		}
	}

	/**
	 * Tests 10000 transactions where each of them stress 1_1 linked persistents objects using 1 thread.
	 */
	public void testRelation1_1ConcurrencyTh10Tx10000() {
		if (!interactive) {
			logger.log(BasicLevel.INFO, "testRelation1_1ConcurrencyTh10Tx10000");
			perform(10, 10000, 10000, 1000000);
		} else {
			logger.log(BasicLevel.INFO, "Non Interactive mode: ignore testRelation1_1ConcurrencyTh10Tx10000");
		}
	}
}
